/* Based on Bare Bones tutorial on OSDev wiki */

#include <stdbool.h>
#include <stddef.h>
#include <stdint.h>

/* Check if the compiler thinks you are targeting the wrong operating system. */
#if defined(__linux__)
#error "You are not using a cross-compiler, you will most certainly run into trouble"
#endif

/* This tutorial will only work for the 32-bit ix86 targets. */
#if !defined(__i386__)
#error "This tutorial needs to be compiled with a ix86-elf compiler"
#endif

/* Hardware text mode color constants. */
enum vga_color {
	VGA_COLOR_BLACK = 0,
	VGA_COLOR_BLUE = 1,
	VGA_COLOR_GREEN = 2,
	VGA_COLOR_CYAN = 3,
	VGA_COLOR_RED = 4,
	VGA_COLOR_MAGENTA = 5,
	VGA_COLOR_BROWN = 6,
	VGA_COLOR_LIGHT_GREY = 7,
	VGA_COLOR_DARK_GREY = 8,
	VGA_COLOR_LIGHT_BLUE = 9,
	VGA_COLOR_LIGHT_GREEN = 10,
	VGA_COLOR_LIGHT_CYAN = 11,
	VGA_COLOR_LIGHT_RED = 12,
	VGA_COLOR_LIGHT_MAGENTA = 13,
	VGA_COLOR_LIGHT_BROWN = 14,
	VGA_COLOR_WHITE = 15,
};

static inline uint8_t vga_entry_color(enum vga_color fg, enum vga_color bg)
{
	return fg | bg << 4;
}

static inline uint16_t vga_entry(unsigned char uc, uint8_t color)
{
	return (uint16_t) uc | (uint16_t) color << 8;
}

size_t strlen(const char* str)
{
	size_t len = 0;
	while (str[len])
		len++;
	return len;
}

static const size_t VGA_WIDTH = 80;
static const size_t VGA_HEIGHT = 25;

size_t terminal_row;
size_t terminal_column;
uint8_t terminal_color;
uint16_t* terminal_buffer;

void terminal_initialize(void)
{
	terminal_row = 0;
	terminal_column = 0;
	terminal_color = vga_entry_color(VGA_COLOR_LIGHT_GREY, VGA_COLOR_BLACK);
	terminal_buffer = (uint16_t*) 0xB8000;
	for (size_t y = 0; y < VGA_HEIGHT; y++) {
		for (size_t x = 0; x < VGA_WIDTH; x++) {
			const size_t index = y * VGA_WIDTH + x;
			terminal_buffer[index] = vga_entry(' ', terminal_color);
		}
	}
}

void terminal_setcolor(uint8_t color)
{
	terminal_color = color;
}

void terminal_putentryat(char c, uint8_t color, size_t x, size_t y)
{
	const size_t index = y * VGA_WIDTH + x;
	terminal_buffer[index] = vga_entry(c, color);
}

void terminal_putchar(char c)
{
	if (c == '\n') {
		terminal_row++;
		terminal_column = 0;
		return;
	}

	terminal_putentryat(c, terminal_color, terminal_column, terminal_row);
	if (++terminal_column == VGA_WIDTH) {
		terminal_column = 0;
		if (++terminal_row == VGA_HEIGHT)
			terminal_row = 0;
	}
}

void terminal_write(const char* data, size_t size)
{
	for (size_t i = 0; i < size; i++)
		terminal_putchar(data[i]);
}

void terminal_writestring(const char* data)
{
	terminal_write(data, strlen(data));
}

// (out|in)port from sortix, Copyright (c) 2014, 2015 Jonas 'Sortie' Termansen.
void outport8(uint16_t port, uint8_t value)
{
	asm volatile ("outb %1, %0" : : "dN" (port), "a" (value));
}

uint8_t inport8(uint16_t port)
{
	uint8_t result;
	asm volatile("inb %1, %0" : "=a" (result) : "dN" (port));
	return result;
}

void command_8042(uint8_t command) {
	outport8(0x64, command);
}

uint8_t status_8042(void) {
	return inport8(0x64);
}

uint8_t read_8042(void) {
	return inport8(0x60);
}

void write_8042(uint8_t data) {
	outport8(0x60, data);
}

bool translation_initially;
bool dual_channel;

uint8_t read_config_8042(void) {
	command_8042(0x20);
	return read_8042();
}

void write_config_8042(uint8_t config) {
	command_8042(0x60);
	write_8042(config);
}

void init_8042(void) {
	// Disable devices
	command_8042(0xad);
	command_8042(0xa7);

	// Flush output buffer
	while (status_8042() & (1<<0))
		(void) read_8042();
	
	// Read configuration
	uint8_t config = read_config_8042();
	translation_initially = config & (1<<6);
	dual_channel = config & (1<<5);

	// Disable IRQ
	config &= ~(1<<0);
	config &= ~(1<<1);
	command_8042(0x60);
	write_8042(config);

	// Self test
	terminal_writestring("8042 self-test... ");
	command_8042(0xaa);
	while (!(status_8042() & (1<<0)));
	if (read_8042() != 0x55)
		terminal_writestring("failed\n");
	else
		terminal_writestring("ok\n");
	// Some controllers need this
	command_8042(0x60);
	write_8042(config);

	// Determine number of channels
	if (dual_channel) {
		command_8042(0xa8);
		dual_channel = !(read_config_8042() & (1<<5));
		if (dual_channel)
			command_8042(0xa7);
	}
	if (dual_channel)
		terminal_writestring("Dual-channel ps/2 controller\n");
	else
		terminal_writestring("Single-channel ps/2 controller\n");

	// Test channels
	terminal_writestring("Testing channel 1... ");
	command_8042(0xab);
	if (read_8042() != 0)
		terminal_writestring("failed\n");
	else
		terminal_writestring("ok\n");
	if (dual_channel) {
		terminal_writestring("Testing channel 2... ");
		command_8042(0xa9);
		if (read_8042() != 0)
			terminal_writestring("failed\n");
		else
			terminal_writestring("ok\n");
	}

	// Enable devices
	command_8042(0xae);
	if (dual_channel)
		command_8042(0xa8);
}

void enable_translation(void) {
	uint8_t config = read_config_8042();
	config |= 1<<6;
	write_config_8042(config);
}

void disable_translation(void) {
	uint8_t config = read_config_8042();
	config &= ~(1<<6);
	write_config_8042(config);
}

void kernel_main(void)
{
	/* Initialize terminal interface */
	terminal_initialize();

	init_8042();
	if (single_channel)

	terminal_writestring("\nHanging\n");
}
